/* 
 * ------------------------------------------------------------------------------
 *
 * MIT License
 *
 * Copyright (c) 2021 Parallel Applications Modelling Group - GMAP
 *      GMAP website: https://gmap.pucrs.br
 *
 * Pontifical Catholic University of Rio Grande do Sul (PUCRS)
 * Av. Ipiranga, 6681, Porto Alegre - Brazil, 90619-900
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * ------------------------------------------------------------------------------
 *
 * The original NPB 3.4 version was written in Fortran and belongs to:
 *      http://www.nas.nasa.gov/Software/NPB/
 *
 * Authors of the Fortran code:
 *      M. Yarrow
 *      C. Kuszmaul
 *
 * ------------------------------------------------------------------------------
 *
 * The serial C++ version is a translation of the original NPB 3.4
 * Serial C++ version: https://github.com/GMAP/NPB-CPP/tree/master/NPB-SER
 *
 * Authors of the C++ code:
 *      Dalvan Griebler <dalvangriebler@gmail.com>
 *      Gabriell Araujo <hexenoften@gmail.com>
 *      Júnior Löff <loffjh@gmail.com>
 *
 * ------------------------------------------------------------------------------
 *
 * The CUDA version is a parallel implementation of the serial C++ version
 * CUDA version: https://github.com/GMAP/NPB-GPU/tree/master/CUDA
 *
 * Authors of the CUDA code:
 *      Gabriell Araujo <hexenoften@gmail.com>
 *
 * ------------------------------------------------------------------------------
 */

#include <skepu>
#include <skepu-lib/io.hpp>
#include "../common/npb-CPP.hpp"
#include "npbparams.hpp"

/*
 * ---------------------------------------------------------------------
 * note: please observe that in the routine conj_grad three 
 * implementations of the sparse matrix-vector multiply have
 * been supplied. the default matrix-vector multiply is not
 * loop unrolled. the alternate implementations are unrolled
 * to a depth of 2 and unrolled to a depth of 8. please
 * experiment with these to find the fastest for your particular
 * architecture. if reporting timing results, any of these three may
 * be used without penalty.
 * ---------------------------------------------------------------------
 * class specific parameters: 
 * it appears here for reference only.
 * these are their values, however, this info is imported in the npbparams.h
 * include file, which is written by the sys/setparams.c program.
 * ---------------------------------------------------------------------
 */
#define NZ (NA*(NONZER+1)*(NONZER+1))
#define NAZ (NA*(NONZER+1))
#define PROFILING_TOTAL_TIME (0)
#define PROFILING_KERNEL_ONE (1)
#define PROFILING_KERNEL_TWO (2)
#define PROFILING_KERNEL_THREE (3)
#define PROFILING_KERNEL_FOUR (4)
#define PROFILING_KERNEL_FIVE (5)
#define PROFILING_KERNEL_SIX (6)
#define PROFILING_KERNEL_SEVEN (7)
#define PROFILING_KERNEL_EIGHT (8)
#define PROFILING_KERNEL_NINE (9)
#define PROFILING_KERNEL_TEN (10)
#define PROFILING_KERNEL_ELEVEN (11)

/* global variables */

static int iv[NA];
static int arow[NA];
static int acol[NAZ];
static double aelt[NAZ];

skepu::Vector<int> colidx(NZ);
skepu::Vector<int> rowstr(NA+1);
skepu::Vector<double> a(NZ);

skepu::Vector<double> x(NA);
skepu::Vector<double> z(NA);
skepu::Vector<double> p(NA);
skepu::Vector<double> q(NA);
skepu::Vector<double> r(NA);

static int naa;
static int nzz;
static int firstrow;
static int lastrow;
static int firstcol;
static int lastcol;
static double amult;
static double tran;

/* function prototypes */
static void conj_grad(
		skepu::Vector<int> &colidx,
		skepu::Vector<int> &rowstr,
		skepu::Vector<double> &x,
		skepu::Vector<double> &z,
		skepu::Vector<double> &a,
		skepu::Vector<double> &p,
		skepu::Vector<double> &q,
		skepu::Vector<double> &r,
		double* rnorm);
static void conj_grad_skepu(double* rnorm);
static void skepu_kernel_one();
static void skepu_kernel_two(double* rho_host);
static void skepu_kernel_three();
static void skepu_kernel_four(double* d_host);
static void skepu_kernel_five(double alpha_host);
static void skepu_kernel_six(double* rho_host);
static void skepu_kernel_seven(double beta_host);
static void skepu_kernel_eight();
static void skepu_kernel_nine(double* sum_host);
static void skepu_kernel_ten(double* norm_temp1, 
		double* norm_temp2);
static void skepu_kernel_eleven(double norm_temp2);
static int icnvrt(double x,
		int ipwr2);
static void makea(int n,
		int nz,
		skepu::Vector<double> &a,
		skepu::Vector<int> &colidx,
		skepu::Vector<int> &rowstr,
		int firstrow,
		int lastrow,
		int firstcol,
		int lastcol,
		int arow[],
		int acol[][NONZER+1],
		double aelt[][NONZER+1],
		int iv[]
	);
static void sparse(
		skepu::Vector<double> &a,
		skepu::Vector<int> &colidx,
		skepu::Vector<int> &rowstr,
		int n,
		int nz,
		int nozer,
		int arow[],
		int acol[][NONZER+1],
		double aelt[][NONZER+1],
		int firstrow,
		int lastrow,
	  int nzloc[],
		double rcond,
		double shift);
static void sprnvc(int n,
		int nz,
		int nn1,
		double v[],
		int iv[]);
static void vecset(int n,
		double v[],
		int iv[],
		int* nzv,
		int i,
		double val);

/* cg */
int main(int argc, char** argv){
#if defined(DO_NOT_ALLOCATE_ARRAYS_WITH_DYNAMIC_MEMORY_AND_AS_SINGLE_DIMENSION)
	printf(" DO_NOT_ALLOCATE_ARRAYS_WITH_DYNAMIC_MEMORY_AND_AS_SINGLE_DIMENSION mode on\n");
#endif
#if defined(PROFILING)
	printf(" PROFILING mode on\n");
#endif
	int	i, j, k, it;
	double zeta;
	double rnorm;
	double norm_temp1, norm_temp2;
	double t, mflops;
	char class_npb;
	boolean verified;
	double zeta_verify_value, epsilon, err;
	
#define NPB_SKEPU_BACKEND "OpenMP"
#define NPB_SKEPU_OMPTHREADS 12
//	skepu::BackendSpec spec{"OpenCL"};
//	skepu::setGlobalBackendSpec(spec);

#define NPB_SKEPU_GPU_THREADS_ONE 1024
#define NPB_SKEPU_GPU_THREADS_TWO 256
#define NPB_SKEPU_GPU_THREADS_THREE 64
#define NPB_SKEPU_GPU_THREADS_FOUR 256
#define NPB_SKEPU_GPU_THREADS_FIVE 64
#define NPB_SKEPU_GPU_THREADS_SIX 256
#define NPB_SKEPU_GPU_THREADS_SEVEN 512
#define NPB_SKEPU_GPU_THREADS_EIGHT 64
#define NPB_SKEPU_GPU_THREADS_NINE 512
#define NPB_SKEPU_GPU_THREADS_TEN 256
#define NPB_SKEPU_GPU_THREADS_ELEVEN 512

	timer_clear(PROFILING_TOTAL_TIME);
#if defined(PROFILING)
	timer_clear(PROFILING_KERNEL_ONE);
	timer_clear(PROFILING_KERNEL_TWO);
	timer_clear(PROFILING_KERNEL_THREE);
	timer_clear(PROFILING_KERNEL_FOUR);
	timer_clear(PROFILING_KERNEL_FIVE);
	timer_clear(PROFILING_KERNEL_SIX);
	timer_clear(PROFILING_KERNEL_SEVEN);
	timer_clear(PROFILING_KERNEL_EIGHT);
	timer_clear(PROFILING_KERNEL_NINE);
	timer_clear(PROFILING_KERNEL_TEN);
	timer_clear(PROFILING_KERNEL_ELEVEN);
#endif

	firstrow = 0;
	lastrow  = NA-1;
	firstcol = 0;
	lastcol  = NA-1;

	if(NA == 1400 && NONZER == 7 && NITER == 15 && SHIFT == 10.0){
		class_npb = 'S';
		zeta_verify_value = 8.5971775078648;
	}else if(NA == 7000 && NONZER == 8 && NITER == 15 && SHIFT == 12.0){
		class_npb = 'W';
		zeta_verify_value = 10.362595087124;
	}else if(NA == 14000 && NONZER == 11 && NITER == 15 && SHIFT == 20.0){
		class_npb = 'A';
		zeta_verify_value = 17.130235054029;
	}else if(NA == 75000 && NONZER == 13 && NITER == 75 && SHIFT == 60.0){
		class_npb = 'B';
		zeta_verify_value = 22.712745482631;
	}else if(NA == 150000 && NONZER == 15 && NITER == 75 && SHIFT == 110.0){
		class_npb = 'C';
		zeta_verify_value = 28.973605592845;
	}else if(NA == 1500000 && NONZER == 21 && NITER == 100 && SHIFT == 500.0){
		class_npb = 'D';
		zeta_verify_value = 52.514532105794;
	}else if(NA == 9000000 && NONZER == 26 && NITER == 100 && SHIFT == 1500.0){
		class_npb = 'E';
		zeta_verify_value = 77.522164599383;
	}else{
		class_npb = 'U';
	}

	printf("\n\n NAS Parallel Benchmarks 4.1 CUDA C++ version - CG Benchmark\n\n");
	printf(" Size: %11d\n", NA);
	printf(" Iterations: %5d\n", NITER);

	naa = NA;
	nzz = NZ;

	/* initialize random number generator */
	tran    = 314159265.0;
	amult   = 1220703125.0;
	zeta    = randlc( &tran, amult );

	makea(naa, 
			nzz, 
			a, 
			colidx, 
			rowstr, 
			firstrow, 
			lastrow, 
			firstcol, 
			lastcol, 
			arow, 
			(int(*)[NONZER+1])(void*)acol, 
			(double(*)[NONZER+1])(void*)aelt,
			iv);

	/*
	 * ---------------------------------------------------------------------
	 * note: as a result of the above call to makea:
	 * values of j used in indexing rowstr go from 0 --> lastrow-firstrow
	 * values of colidx which are col indexes go from firstcol --> lastcol
	 * so:
	 * shift the col index vals from actual (firstcol --> lastcol) 
	 * to local, i.e., (0 --> lastcol-firstcol)
	 * ---------------------------------------------------------------------
	 */
	for(j = 0; j < lastrow - firstrow + 1; j++){
		for(k = rowstr[j]; k < rowstr[j+1]; k++){
			colidx[k] = colidx[k] - firstcol;
		}
	}

	/* set starting vector to (1, 1, .... 1) */
	for(i = 0; i < NA+1; i++){
		x[i] = 1.0;
	}
	for(j = 0; j<lastcol-firstcol+1; j++){
		q[j] = 0.0;
		z[j] = 0.0;
		r[j] = 0.0;
		p[j] = 0.0;
	}
	zeta = 0.0;

	/*
	 * -------------------------------------------------------------------
	 * ---->
	 * do one iteration untimed to init all code and data page tables
	 * ----> (then reinit, start timing, to niter its)
	 * -------------------------------------------------------------------*/
	for(it = 1; it <= 1; it++){
		/* the call to the conjugate gradient routine */
		conj_grad(colidx, rowstr, x, z, a, p, q, r, &rnorm);

		/*
		 * --------------------------------------------------------------------
		 * zeta = shift + 1/(x.z)
		 * so, first: (x.z)
		 * also, find norm of z
		 * so, first: (z.z)
		 * --------------------------------------------------------------------
		 */
		norm_temp1 = 0.0;
		norm_temp2 = 0.0;
		for(j = 0; j < lastcol - firstcol + 1; j++){
			norm_temp1 = norm_temp1 + x[j] * z[j];
			norm_temp2 = norm_temp2 + z[j] * z[j];
		}
		norm_temp2 = 1.0 / sqrt(norm_temp2);

		/* normalize z to obtain x */
		for(j = 0; j < lastcol - firstcol + 1; j++){     
			x[j] = norm_temp2 * z[j];
		}
	} /* end of do one iteration untimed */

	/* set starting vector to (1, 1, .... 1) */	
	for(i = 0; i < NA+1; i++){
		x[i] = 1.0;
	}
	zeta = 0.0;

	timer_start(PROFILING_TOTAL_TIME);

	/*
	 * --------------------------------------------------------------------
	 * ---->
	 * main iteration for inverse power method
	 * ---->
	 * --------------------------------------------------------------------
	 */
	for(it = 1; it <= NITER; it++){
		/* the call to the conjugate gradient routine */
		conj_grad_skepu(&rnorm);

		/*
		 * --------------------------------------------------------------------
		 * zeta = shift + 1/(x.z)
		 * so, first: (x.z)
		 * also, find norm of z
		 * so, first: (z.z)
		 * --------------------------------------------------------------------
		 */
		skepu_kernel_ten(&norm_temp1, &norm_temp2);
		norm_temp2 = 1.0 / sqrt(norm_temp2);
		zeta = SHIFT + 1.0 / norm_temp1;
		if(it==1){printf("\n   iteration           ||r||                 zeta\n");}
		printf("    %5d       %20.14e%20.13e\n", it, rnorm, zeta);

		/* normalize z to obtain x */
		skepu_kernel_eleven(norm_temp2);
	} /* end of main iter inv pow meth */

	timer_stop(PROFILING_TOTAL_TIME);

	/*
	 * --------------------------------------------------------------------
	 * end of timed section
	 * --------------------------------------------------------------------
	 */

	t = timer_read(PROFILING_TOTAL_TIME);

	printf(" Benchmark completed\n");

	epsilon = 1.0e-10;
	if(class_npb != 'U'){
		err = fabs(zeta - zeta_verify_value) / zeta_verify_value;
		if(err <= epsilon){
			verified = TRUE;
			printf(" VERIFICATION SUCCESSFUL\n");
			printf(" Zeta is    %20.13e\n", zeta);
			printf(" Error is   %20.13e\n", err);
		}else{
			verified = FALSE;
			printf(" VERIFICATION FAILED\n");
			printf(" Zeta                %20.13e\n", zeta);
			printf(" The correct zeta is %20.13e\n", zeta_verify_value);
		}
	}else{
		verified = FALSE;
		printf(" Problem size unknown\n");
		printf(" NO VERIFICATION PERFORMED\n");
	}
	if(t != 0.0){
		mflops = (double)(2.0*NITER*NA)
			* (3.0+(double)(NONZER*(NONZER+1))
					+ 25.0
					* (5.0+(double)(NONZER*(NONZER+1)))+3.0)
			/ t / 1000000.0;
	}else{
		mflops = 0.0;
	}

	char gpu_config[256];
	char gpu_config_string[2048];
#if defined(PROFILING)
	sprintf(gpu_config, "%5s\t%25s\t%25s\t%25s\n", "GPU Kernel", "Threads Per Block", "Time in Seconds", "Time in Percentage");
	strcpy(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " one", threads_per_block_on_kernel_one, timer_read(PROFILING_KERNEL_ONE), (timer_read(PROFILING_KERNEL_ONE)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " two", threads_per_block_on_kernel_two, timer_read(PROFILING_KERNEL_TWO), (timer_read(PROFILING_KERNEL_TWO)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " three", threads_per_block_on_kernel_three, timer_read(PROFILING_KERNEL_THREE), (timer_read(PROFILING_KERNEL_THREE)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " four", threads_per_block_on_kernel_four, timer_read(PROFILING_KERNEL_FOUR), (timer_read(PROFILING_KERNEL_FOUR)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " five", threads_per_block_on_kernel_five, timer_read(PROFILING_KERNEL_FIVE), (timer_read(PROFILING_KERNEL_FIVE)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " six", threads_per_block_on_kernel_six, timer_read(PROFILING_KERNEL_SIX), (timer_read(PROFILING_KERNEL_SIX)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " seven", threads_per_block_on_kernel_seven, timer_read(PROFILING_KERNEL_SEVEN), (timer_read(PROFILING_KERNEL_SEVEN)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " eight", threads_per_block_on_kernel_eight, timer_read(PROFILING_KERNEL_EIGHT), (timer_read(PROFILING_KERNEL_EIGHT)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " nine", threads_per_block_on_kernel_nine, timer_read(PROFILING_KERNEL_NINE), (timer_read(PROFILING_KERNEL_NINE)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " ten", threads_per_block_on_kernel_ten, timer_read(PROFILING_KERNEL_TEN), (timer_read(PROFILING_KERNEL_TEN)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " eleven", threads_per_block_on_kernel_eleven, timer_read(PROFILING_KERNEL_ELEVEN), (timer_read(PROFILING_KERNEL_ELEVEN)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
#endif

	c_print_results((char*)"CG",
			class_npb,
			NA,
			0,
			0,
			NITER,
			t,
			mflops,
			(char*)"          floating point",
			verified,
			(char*)NPBVERSION,
			(char*)COMPILETIME,
			(char*)COMPILERVERSION,
			(char*)CS1,
			(char*)CS2,
			(char*)CS3,
			(char*)CS4,
			(char*)CS5,
			(char*)CS6,
			(char*)CS7);

	return 0;
}

/*
 * ---------------------------------------------------------------------
 * floating point arrays here are named as in NPB1 spec discussion of 
 * CG algorithm
 * ---------------------------------------------------------------------
 */
static void conj_grad(
		skepu::Vector<int> &colidx,
		skepu::Vector<int> &rowstr,
		skepu::Vector<double> &x,
		skepu::Vector<double> &z,
		skepu::Vector<double> &a,
		skepu::Vector<double> &p,
		skepu::Vector<double> &q,
		skepu::Vector<double> &r,
		double* rnorm
	){
	int j, k;
	int cgit, cgitmax;
	double d, sum, rho, rho0, alpha, beta;

	cgitmax = 25;

	rho = 0.0;

	/* initialize the CG algorithm */
	for(j = 0; j < naa+1; j++){
		q[j] = 0.0;
		z[j] = 0.0;
		r[j] = x[j];
		p[j] = r[j];
	}

	/*
	 * --------------------------------------------------------------------
	 * rho = r.r
	 * now, obtain the norm of r: First, sum squares of r elements locally...
	 * --------------------------------------------------------------------
	 */
	for(j = 0; j < lastcol - firstcol + 1; j++){
		rho = rho + r[j]*r[j];
	}

	/* the conj grad iteration loop */
	for(cgit = 1; cgit <= cgitmax; cgit++){
		/*
		 * ---------------------------------------------------------------------
		 * q = A.p
		 * the partition submatrix-vector multiply: use workspace w
		 * ---------------------------------------------------------------------
		 * 
		 * note: this version of the multiply is actually (slightly: maybe %5) 
		 * faster on the sp2 on 16 nodes than is the unrolled-by-2 version 
		 * below. on the Cray t3d, the reverse is TRUE, i.e., the 
		 * unrolled-by-two version is some 10% faster.  
		 * the unrolled-by-8 version below is significantly faster
		 * on the Cray t3d - overall speed of code is 1.5 times faster.
		 */
		for(j = 0; j < lastrow - firstrow + 1; j++){
			sum = 0.0;
			for(k = rowstr[j]; k < rowstr[j+1]; k++){
				sum = sum + a[k]*p[colidx[k]];
			}
			q[j] = sum;
		}

		/*
		 * --------------------------------------------------------------------
		 * obtain p.q
		 * --------------------------------------------------------------------
		 */
		d = 0.0;
		for (j = 0; j < lastcol - firstcol + 1; j++) {
			d = d + p[j]*q[j];
		}

		/*
		 * --------------------------------------------------------------------
		 * obtain alpha = rho / (p.q)
		 * -------------------------------------------------------------------
		 */
		alpha = rho / d;

		/*
		 * --------------------------------------------------------------------
		 * save a temporary of rho
		 * --------------------------------------------------------------------
		 */
		rho0 = rho;

		/*
		 * ---------------------------------------------------------------------
		 * obtain z = z + alpha*p
		 * and    r = r - alpha*q
		 * ---------------------------------------------------------------------
		 */
		rho = 0.0;
		for(j = 0; j < lastcol - firstcol + 1; j++){
			z[j] = z[j] + alpha*p[j];
			r[j] = r[j] - alpha*q[j];
		}

		/*
		 * ---------------------------------------------------------------------
		 * rho = r.r
		 * now, obtain the norm of r: first, sum squares of r elements locally...
		 * ---------------------------------------------------------------------
		 */
		for(j = 0; j < lastcol - firstcol + 1; j++){
			rho = rho + r[j]*r[j];
		}

		/*
		 * ---------------------------------------------------------------------
		 * obtain beta
		 * ---------------------------------------------------------------------
		 */
		beta = rho / rho0;

		/*
		 * ---------------------------------------------------------------------
		 * p = r + beta*p
		 * ---------------------------------------------------------------------
		 */
		for(j = 0; j < lastcol - firstcol + 1; j++){
			p[j] = r[j] + beta*p[j];
		}
	} /* end of do cgit=1, cgitmax */

	/*
	 * ---------------------------------------------------------------------
	 * compute residual norm explicitly: ||r|| = ||x - A.z||
	 * first, form A.z
	 * the partition submatrix-vector multiply
	 * ---------------------------------------------------------------------
	 */
	sum = 0.0;
	for(j = 0; j < lastrow - firstrow + 1; j++){
		d = 0.0;
		for(k = rowstr[j]; k < rowstr[j+1]; k++){
			d = d + a[k]*z[colidx[k]];
		}
		r[j] = d;
	}

	/*
	 * ---------------------------------------------------------------------
	 * at this point, r contains A.z
	 * ---------------------------------------------------------------------
	 */
	for(j = 0; j < lastcol-firstcol+1; j++){
		d   = x[j] - r[j];
		sum = sum + d*d;
	}

	*rnorm = sqrt(sum);
}

static void conj_grad_skepu(double* rnorm){
	double d, sum, rho, rho0, alpha, beta;
	int cgit, cgitmax = 25;

	/* initialize the CG algorithm */
	skepu_kernel_one();

	/* rho = r.r - now, obtain the norm of r: first, sum squares of r elements locally */
	skepu_kernel_two(&rho);

	/* the conj grad iteration loop */
	for(cgit = 1; cgit <= cgitmax; cgit++){
		/* q = A.p */
		skepu_kernel_three();

		/* obtain p.q */
		skepu_kernel_four(&d);

		alpha = rho / d;

		/* save a temporary of rho */
		rho0 = rho;

		/* obtain (z = z + alpha*p) and (r = r - alpha*q) */
		skepu_kernel_five(alpha);

		/* rho = r.r - now, obtain the norm of r: first, sum squares of r elements locally */
		skepu_kernel_six(&rho);

		/* obtain beta */
		beta = rho / rho0;

		/* p = r + beta*p */
		skepu_kernel_seven(beta);
	} /* end of do cgit=1, cgitmax */

	/* compute residual norm explicitly:  ||r|| = ||x - A.z|| */
	skepu_kernel_eight();

	/* at this point, r contains A.z */
	skepu_kernel_nine(&sum);

	*rnorm = sqrt(sum);
}




static void skepu_kernel_one(){   
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_ONE);
#endif
	
	auto skepu_skel_one = skepu::Map([](double x) -> skepu::multiple<double, double, double, double>
	{
		return skepu::ret(0.0, 0.0, x, x);
	});
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_ONE);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_one.setBackend(spec);
	
	skepu_skel_one(q, z, r, p, x);
	
//	skepu::io::cout << "PRINT: r\n" << r << "\n\n";
//	skepu::io::cout << "PRINT: p\n" << p << "\n\n";
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_ONE);
#endif
}

static void skepu_kernel_two(double* rho_host){  
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_TWO);
#endif
	
	auto skepu_skel_two = skepu::MapReduce(
		[](double r)           -> double { return r * r; }, // map
		[](double l, double r) -> double { return l + r; }  // reduce
	);
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_TWO);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_two.setBackend(spec);
	
	*rho_host = skepu_skel_two(r);
	
//	skepu::io::cout << "PRINT rho\n" << *rho_host << "\n\n";
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_TWO);
#endif
}

static void skepu_kernel_three(){
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_THREE);
#endif
	
	auto skepu_skel_three = skepu::Map([](
		skepu::Index1D     index,
		skepu::Vec<int>    colidx,
		skepu::Vec<int>    rowstr,
		skepu::Vec<double> a,
		skepu::Vec<double> p) -> double
	{
		int begin = rowstr(index.i);
		int end   = rowstr(index.i + 1);
		double sum = 0.0;
		for (int k = begin; k < end; ++k)
		{
			sum += a(k) * p(colidx(k));
		}
		return sum;
	});
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_THREE);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_three.setBackend(spec);
	
	skepu_skel_three(q, colidx, rowstr, a, p);
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_THREE);
#endif
}

static void skepu_kernel_four(double* d_host){   
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_FOUR);
#endif
	
	auto skepu_skel_four = skepu::MapReduce(
		[](double p, double q) -> double { return p * q; }, // map
		[](double l, double r) -> double { return l + r; }  // reduce
	);
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_FOUR);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_four.setBackend(spec);
	
	*d_host = skepu_skel_four(p, q);
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_FOUR);
#endif
}

static void skepu_kernel_five(double alpha_host){
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_FIVE);
#endif
	
	auto skepu_skel_five = skepu::Map<4>([]
		(double p, double z, double q, double r, double alpha)
		-> skepu::multiple<double, double>
	{
		double new_z = z + alpha * p;
		double new_r = r - alpha * q;
		return skepu::ret(new_z, new_r);
	});
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_FIVE);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_five.setBackend(spec);
	
	skepu_skel_five(z, r, p, z, q, r, alpha_host); // note: z, r are both inputs and outputs
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_FIVE);
#endif
}

static void skepu_kernel_six(double* rho_host){
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_SIX);
#endif
	
	auto skepu_skel_six = skepu::MapReduce(
		[](double r)           -> double { return r * r; }, // map
		[](double l, double r) -> double { return l + r; }  // reduce
	);
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_SIX);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_six.setBackend(spec);
	
	*rho_host = skepu_skel_six(r);
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_SIX);
#endif
}

static void skepu_kernel_seven(double beta_host){
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_SEVEN);
#endif
	
	auto skepu_skel_seven = skepu::Map<2>([](double p, double r, double beta) -> double
	{
		return r + beta * p;
	});
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_SEVEN);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_seven.setBackend(spec);
	
	skepu_skel_seven(p, p, r, beta_host); // note: p both input and output
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_SEVEN);
#endif
}

static void skepu_kernel_eight(){
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_EIGHT);
#endif
	
	/* compute residual norm explicitly:  ||r|| = ||x - A.z|| */
	auto skepu_skel_eight = skepu::Map([](
		skepu::Index1D     index,
		skepu::Vec<int>    colidx,
		skepu::Vec<int>    rowstr,
		skepu::Vec<double> a,
		skepu::Vec<double> z) -> double
	{
		int begin = rowstr(index.i);
		int end   = rowstr(index.i + 1);
		double sum = 0.0;
		for (int k = begin; k < end; ++k)
		{
			sum += a(k) * z(colidx(k));
		}
		return sum;
	});
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_EIGHT);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_eight.setBackend(spec);
	
	skepu_skel_eight(r, colidx, rowstr, a, z);
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_EIGHT);
#endif
}

static void skepu_kernel_nine(double* sum_host){ 
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_NINE);
#endif
	
	auto skepu_skel_nine = skepu::MapReduce(
		[](double r, double x) -> double { double tmp = x - r; return tmp * tmp; }, // map
		[](double l, double r) -> double { return l + r; }  // reduce
	);
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_NINE);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_nine.setBackend(spec);
	
	*sum_host = skepu_skel_nine(r, x);
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_NINE);
#endif
}

static void skepu_kernel_ten(double* norm_temp1, 
		double* norm_temp2){
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_TEN);
#endif
	
	auto skepu_skel_ten_1 = skepu::MapReduce(
		[](double x, double z) -> double { return x * z; }, // map
		[](double l, double r) -> double { return l + r; }  // reduce
	);
	
	auto skepu_skel_ten_2 = skepu::MapReduce(
		[](double z)           -> double { return z * z; }, // map
		[](double l, double r) -> double { return l + r; }  // reduce
	);
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_TEN);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_ten_1.setBackend(spec);
	skepu_skel_ten_2.setBackend(spec);
	
	*norm_temp1 = skepu_skel_ten_1(x, z);
	*norm_temp2 = skepu_skel_ten_2(z);
	
	/* // Alternative, in a future version of SkePU:
	auto skepu_skel_ten = skepu::MapReduce(
		[](double x, double z) -> skepu::multiple<double, double> { return skepu::ret(x * z, z * z); }, // map
		[](double l, double r) -> double { return l + r; }  // reduce
	);
	
	std::tie(*norm_temp1, *norm_temp2) = skepu_skel_ten(x, z);
	*/
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_TEN);
#endif
}

static void skepu_kernel_eleven(double norm_temp2){   
#if defined(PROFILING)
	timer_start(PROFILING_KERNEL_ELEVEN);
#endif
	
	auto skepu_skel_eleven = skepu::Map<1>([](double z, double norm_temp2) -> double
	{
		return norm_temp2 * z;
	});
	
	skepu::BackendSpec spec{NPB_SKEPU_BACKEND};
	spec.setGPUThreads(NPB_SKEPU_GPU_THREADS_ELEVEN);
	spec.setCPUThreads(NPB_SKEPU_OMPTHREADS);
	skepu_skel_eleven.setBackend(spec);
	
	skepu_skel_eleven(x, z, norm_temp2);
	
#if defined(PROFILING)
	timer_stop(PROFILING_KERNEL_ELEVEN);
#endif
}

/*
 * ---------------------------------------------------------------------
 * scale a double precision number x in (0,1) by a power of 2 and chop it
 * ---------------------------------------------------------------------
 */
static int icnvrt(double x, int ipwr2){
	return (int)(ipwr2 * x);
}

/*
 * ---------------------------------------------------------------------
 * generate the test problem for benchmark 6
 * makea generates a sparse matrix with a
 * prescribed sparsity distribution
 *
 * parameter    type        usage
 *
 * input
 *
 * n            i           number of cols/rows of matrix
 * nz           i           nonzeros as declared array size
 * rcond        r*8         condition number
 * shift        r*8         main diagonal shift
 *
 * output
 *
 * a            r*8         array for nonzeros
 * colidx       i           col indices
 * rowstr       i           row pointers
 *
 * workspace
 *
 * iv, arow, acol i
 * aelt           r*8
 * ---------------------------------------------------------------------
 */
static void makea(
		// int n,
		// int nz,
		// double a[],
		// int colidx[],
		// int rowstr[],
		// int firstrow,
		// int lastrow,
		// int firstcol,
		// int lastcol,
		// int arow[],
		// int acol[][NONZER+1],
		// double aelt[][NONZER+1],
		// int iv[]
		int n,
		int nz,
		skepu::Vector<double> &a,
		skepu::Vector<int> &colidx,
		skepu::Vector<int> &rowstr,
		int firstrow,
		int lastrow,
		int firstcol,
		int lastcol,
		int arow[],
		int acol[][NONZER+1],
		double aelt[][NONZER+1],
		int iv[]
	){
	int iouter, ivelt, nzv, nn1;
	int ivc[NONZER+1];
	double vc[NONZER+1];

	/*
	 * --------------------------------------------------------------------
	 * nonzer is approximately  (int(sqrt(nnza /n)));
	 * --------------------------------------------------------------------
	 * nn1 is the smallest power of two not less than n
	 * --------------------------------------------------------------------
	 */
	nn1 = 1;
	do{
		nn1 = 2 * nn1;
	}while(nn1 < n);

	/*
	 * -------------------------------------------------------------------
	 * generate nonzero positions and save for the use in sparse
	 * -------------------------------------------------------------------
	 */
	for(iouter = 0; iouter < n; iouter++){
		nzv = NONZER;
		sprnvc(n, nzv, nn1, vc, ivc);
		vecset(n, vc, ivc, &nzv, iouter+1, 0.5);
		arow[iouter] = nzv;
		for(ivelt = 0; ivelt < nzv; ivelt++){
			acol[iouter][ivelt] = ivc[ivelt] - 1;
			aelt[iouter][ivelt] = vc[ivelt];
		}
	}

	/*
	 * ---------------------------------------------------------------------
	 * ... make the sparse matrix from list of elements with duplicates
	 * (iv is used as  workspace)
	 * ---------------------------------------------------------------------
	 */
	sparse(a,
			colidx,
			rowstr,
			n,
			nz,
			NONZER,
			arow,
			acol,
			aelt,
			firstrow,
			lastrow,
			iv,
			RCOND,
			SHIFT);
}

/*
 * ---------------------------------------------------------------------
 * rows range from firstrow to lastrow
 * the rowstr pointers are defined for nrows = lastrow-firstrow+1 values
 * ---------------------------------------------------------------------
 */
static void sparse(
		skepu::Vector<double> &a,
		skepu::Vector<int> &colidx,
		skepu::Vector<int> &rowstr,
		int n,
		int nz,
		int nozer,
	//	skepu::Vector<int> &arow,
	//	skepu::Matrix<int> &acol,
	//	skepu::Matrix<double> &aelt,
		int arow[],
		int acol[][NONZER+1],
		double aelt[][NONZER+1],
		int firstrow,
		int lastrow,
	//	skepu::Vector<int> &nzloc,
	  int nzloc[],
		double rcond,
		double shift
	){	
	int nrows;

	/*
	 * ---------------------------------------------------
	 * generate a sparse matrix from a list of
	 * [col, row, element] tri
	 * ---------------------------------------------------
	 */
	int i, j, j1, j2, nza, k, kk, nzrow, jcol;
	double size, scale, ratio, va;
	boolean goto_40;

	/*
	 * --------------------------------------------------------------------
	 * how many rows of result
	 * --------------------------------------------------------------------
	 */
	nrows = lastrow - firstrow + 1;

	/*
	 * --------------------------------------------------------------------
	 * ...count the number of triples in each row
	 * --------------------------------------------------------------------
	 */
	for(j = 0; j < nrows+1; j++){
		rowstr[j] = 0;
	}
	for(i = 0; i < n; i++){
		for(nza = 0; nza < arow[i]; nza++){
			j = acol[i][nza] + 1;
			rowstr[j] = rowstr[j] + arow[i];
		}
	}
	rowstr[0] = 0;
	for(j = 1; j < nrows+1; j++){
		rowstr[j] = rowstr[j] + rowstr[j-1];
	}
	nza = rowstr[nrows] - 1;

	/*
	 * ---------------------------------------------------------------------
	 * ... rowstr(j) now is the location of the first nonzero
	 * of row j of a
	 * ---------------------------------------------------------------------
	 */
	if(nza > nz){
		printf("Space for matrix elements exceeded in sparse\n");
		printf("nza, nzmax = %d, %d\n", nza, nz);
		exit(EXIT_FAILURE);
	}

	/*
	 * ---------------------------------------------------------------------
	 * ... preload data pages
	 * ---------------------------------------------------------------------
	 */
	for(j = 0; j < nrows; j++){
		for(k = rowstr[j]; k < rowstr[j+1]; k++){
			a[k] = 0.0;
			colidx[k] = -1;
		}
		nzloc[j] = 0;
	}

	/*
	 * ---------------------------------------------------------------------
	 * ... generate actual values by summing duplicates
	 * ---------------------------------------------------------------------
	 */
	size = 1.0;
	ratio = pow(rcond, (1.0 / (double)(n)));
	for(i = 0; i < n; i++){
		for(nza = 0; nza < arow[i]; nza++){
			j = acol[i][nza];

			scale = size * aelt[i][nza];
			for(nzrow = 0; nzrow < arow[i]; nzrow++){
				jcol = acol[i][nzrow];
				va = aelt[i][nzrow] * scale;

				/*
				 * --------------------------------------------------------------------
				 * ... add the identity * rcond to the generated matrix to bound
				 * the smallest eigenvalue from below by rcond
				 * --------------------------------------------------------------------
				 */
				if(jcol == j && j == i){
					va = va + rcond - shift;
				}

				goto_40 = FALSE;
				for(k = rowstr[j]; k < rowstr[j+1]; k++){
					if(colidx[k] > jcol){
						/*
						 * ----------------------------------------------------------------
						 * ... insert colidx here orderly
						 * ----------------------------------------------------------------
						 */
						for(kk = rowstr[j+1]-2; kk >= k; kk--){
							if(colidx[kk] > -1){
								a[kk+1] = a[kk];
								colidx[kk+1] = colidx[kk];
							}
						}
						colidx[k] = jcol;
						a[k]  = 0.0;
						goto_40 = TRUE;
						break;
					}else if(colidx[k] == -1){
						colidx[k] = jcol;
						goto_40 = TRUE;
						break;
					}else if(colidx[k] == jcol){
						/*
						 * --------------------------------------------------------------
						 * ... mark the duplicated entry
						 * -------------------------------------------------------------
						 */
						nzloc[j] = nzloc[j] + 1;
						goto_40 = TRUE;
						break;
					}
				}
				if(goto_40 == FALSE){
					printf("internal error in sparse: i=%d\n", i);
					exit(EXIT_FAILURE);
				}
				a[k] = a[k] + va;
			}
		}
		size = size * ratio;
	}

	/*
	 * ---------------------------------------------------------------------
	 * ... remove empty entries and generate final results
	 * ---------------------------------------------------------------------
	 */
	for(j = 1; j < nrows; j++){
		nzloc[j] = nzloc[j] + nzloc[j-1];
	}

	for(j = 0; j < nrows; j++){
		if(j > 0){
			j1 = rowstr[j] - nzloc[j-1];
		}else{
			j1 = 0;
		}
		j2 = rowstr[j+1] - nzloc[j];
		nza = rowstr[j];
		for(k = j1; k < j2; k++){
			a[k] = a[nza];
			colidx[k] = colidx[nza];
			nza = nza + 1;
		}
	}
	for(j = 1; j < nrows+1; j++){
		rowstr[j] = rowstr[j] - nzloc[j-1];
	}
	nza = rowstr[nrows] - 1;
}

/*
 * ---------------------------------------------------------------------
 * generate a sparse n-vector (v, iv)
 * having nzv nonzeros
 *
 * mark(i) is set to 1 if position i is nonzero.
 * mark is all zero on entry and is reset to all zero before exit
 * this corrects a performance bug found by John G. Lewis, caused by
 * reinitialization of mark on every one of the n calls to sprnvc
 * ---------------------------------------------------------------------
 */
static void sprnvc(int n, int nz, int nn1, double v[], int iv[]){
	int nzv, ii, i;
	double vecelt, vecloc;

	nzv = 0;

	while(nzv < nz){
		vecelt = randlc(&tran, amult);

		/*
		 * --------------------------------------------------------------------
		 * generate an integer between 1 and n in a portable manner
		 * --------------------------------------------------------------------
		 */
		vecloc = randlc(&tran, amult);
		i = icnvrt(vecloc, nn1) + 1;
		if(i>n){continue;}

		/*
		 * --------------------------------------------------------------------
		 * was this integer generated already?
		 * --------------------------------------------------------------------
		 */
		boolean was_gen = FALSE;
		for(ii = 0; ii < nzv; ii++){
			if(iv[ii] == i){
				was_gen = TRUE;
				break;
			}
		}
		if(was_gen){continue;}
		v[nzv] = vecelt;
		iv[nzv] = i;
		nzv = nzv + 1;
	}
}

/*
 * --------------------------------------------------------------------
 * set ith element of sparse vector (v, iv) with
 * nzv nonzeros to val
 * --------------------------------------------------------------------
 */
static void vecset(int n, double v[], int iv[], int* nzv, int i, double val){
	int k;
	boolean set;

	set = FALSE;
	for(k = 0; k < *nzv; k++){
		if(iv[k] == i){
			v[k] = val;
			set  = TRUE;
		}
	}
	if(set == FALSE){
		v[*nzv]  = val;
		iv[*nzv] = i;
		*nzv     = *nzv + 1;
	}
}
