/* 
 * ------------------------------------------------------------------------------
 *
 * MIT License
 *
 * Copyright (c) 2021 Parallel Applications Modelling Group - GMAP
 *      GMAP website: https://gmap.pucrs.br
 *
 * Pontifical Catholic University of Rio Grande do Sul (PUCRS)
 * Av. Ipiranga, 6681, Porto Alegre - Brazil, 90619-900
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * ------------------------------------------------------------------------------
 *
 * The original NPB 3.4 version was written in Fortran and belongs to:
 *      http://www.nas.nasa.gov/Software/NPB/
 *
 * Authors of the Fortran code:
 *      P. O. Frederickson
 *      D. H. Bailey
 *      A. C. Woo
 *
 * ------------------------------------------------------------------------------
 *
 * The serial C++ version is a translation of the original NPB 3.4
 * Serial C++ version: https://github.com/GMAP/NPB-CPP/tree/master/NPB-SER
 *
 * Authors of the C++ code:
 *      Dalvan Griebler <dalvangriebler@gmail.com>
 *      Gabriell Araujo <hexenoften@gmail.com>
 *      Júnior Löff <loffjh@gmail.com>
 *
 * ------------------------------------------------------------------------------
 *
 * The CUDA version is a parallel implementation of the serial C++ version
 * CUDA version: https://github.com/GMAP/NPB-GPU/tree/master/CUDA
 *
 * Authors of the CUDA code:
 *      Gabriell Araujo <hexenoften@gmail.com>
 *
 * ------------------------------------------------------------------------------
 */

#include <skepu>
#include "../common/npb-CPP.hpp"
#include "npbparams.hpp"

#define	MK (16)
#define	MM (M - MK)
#define	NN (1 << MM)
#define	NK (1 << MK)
#define	NQ (10)
#define EPSILON (1.0e-8)
#define	A (1220703125.0)
#define	S (271828183.0)
#define NK_PLUS ((2*NK)+1)
#define RECOMPUTATION (128)
#define PROFILING_TOTAL_TIME (0)

/* global variables */
static double q[NQ];

skepu::Vector<double> q0(NN);
skepu::Vector<double> q1(NN);
skepu::Vector<double> q2(NN);
skepu::Vector<double> q3(NN);
skepu::Vector<double> q4(NN);
skepu::Vector<double> q5(NN);
skepu::Vector<double> q6(NN);
skepu::Vector<double> q7(NN);
skepu::Vector<double> q8(NN);
skepu::Vector<double> q9(NN);
skepu::Vector<double> sx_skepu(NN);
skepu::Vector<double> sy_skepu(NN);


double randlc_skepu(double* x, double a)
{
	double t1,t2,t3,t4,a1,a2,x1,x2,z;
	t1 = R23 * a;
	a1 = (int)t1;
	a2 = a - T23 * a1;
	t1 = R23 * (*x);
	x1 = (int)t1;
	x2 = (*x) - T23 * x1;
	t1 = a1 * x2 + a2 * x1;
	t2 = (int)(R23 * t1);
	z = t1 - T23 * t2;
	t3 = T23 * z + a2 * x2;
	t4 = (int)(R46 * t3);
	(*x) = t3 - T46 * t4;
	return (R46 * (*x));
}

void vranlc_skepu(
	int n, 
	double* x_seed, 
	double a, 
	double* y)
{
	int i;
	double x,t1,t2,t3,t4,a1,a2,x1,x2,z;
	t1 = R23 * a;
	a1 = (int)t1;
	a2 = a - T23 * a1;
	x = *x_seed;
	for(i=0; i<n; i++){
		t1 = R23 * x;
		x1 = (int)t1;
		x2 = x - T23 * x1;
		t1 = a1 * x2 + a2 * x1;
		t2 = (int)(R23 * t1);
		z = t1 - T23 * t2;
		t3 = T23 * z + a2 * x2;
		t4 = (int)(R46 * t3);
		x = t3 - T46 * t4;
		y[i] = R46 * x;
	}
	*x_seed = x;
}

auto mapfun(skepu::Index1D index, double an) -> skepu::multiple<double, double, double, double, double, double, double, double, double, double, double, double>
{	
	double x_local[2*RECOMPUTATION];
	double q_local[NQ]; 
	double sx_local = 0.0, sy_local = 0.0;
	double t3, t4, x1, x2, seed;
	int kk = index.i;
	double t1=S;
	double t2=an;

	for (int i=0; i<NQ; i++){
		q_local[i]=0.0;
	}

	/* find starting seed t1 for this kk */
	for(int i=1; i<=100; i++){
		int ik=kk/2;
		if((2*ik)!=kk){t3=randlc_skepu(&t1, t2);}
		if(ik==0){break;}
		t3=randlc_skepu(&t2, t2);
		kk=ik;
	} 

	seed=t1;
	for(int ii=0; ii<NK; ii=ii+RECOMPUTATION){
		/* compute uniform pseudorandom numbers */
		vranlc_skepu(2*RECOMPUTATION, &seed, A, x_local);

		/*
		 * compute gaussian deviates by acceptance-rejection method and
		 * tally counts in concentric square annuli. this loop is not
		 * vectorizable.
		 */
		for(int i=0; i<RECOMPUTATION; i++){
			x1=2.0*x_local[2*i]-1.0;
			x2=2.0*x_local[2*i+1]-1.0;
			t1=x1*x1+x2*x2;
			if(t1<=1.0){
				t2=sqrt(-2.0*log(t1)/t1);
				t3=(x1*t2);
				t4=(x2*t2);
				int l=max(fabs(t3), fabs(t4));
				q_local[l]+=1.0;
				sx_local+=t3;
				sy_local+=t4;
			}
		}
	}
	
	return skepu::ret(
		q_local[0],
		q_local[1],
		q_local[2],
		q_local[3],
		q_local[4],
		q_local[5],
		q_local[6],
		q_local[7],
		q_local[8],
		q_local[9],
		sx_local,
		sy_local);
}

double redfun(double lhs, double rhs)
{
	return lhs + rhs;
}



/* ep */
int main(int argc, char** argv){
#if defined(DO_NOT_ALLOCATE_ARRAYS_WITH_DYNAMIC_MEMORY_AND_AS_SINGLE_DIMENSION)
	printf(" DO_NOT_ALLOCATE_ARRAYS_WITH_DYNAMIC_MEMORY_AND_AS_SINGLE_DIMENSION mode on\n");
#endif
#if defined(PROFILING)
	printf(" PROFILING mode on\n");
#endif
	double Mops, t1;
	double sx, sy, tm, an, gc;
	double sx_verify_value, sy_verify_value, sx_err, sy_err;
	int i, j, nit, block;
	boolean verified;
	char size[16];

	/*
	 * --------------------------------------------------------------------
	 * because the size of the problem is too large to store in a 32-bit
	 * integer for some classes, we put it into a string (for printing).
	 * have to strip off the decimal point put in there by the floating
	 * point print statement (internal file)
	 * --------------------------------------------------------------------
	 */    
	sprintf(size, "%15.0f", pow(2.0, M+1));
	j = 14;
	if(size[j]=='.'){j--;}
	size[j+1] = '\0';
	printf("\n\n NAS Parallel Benchmarks 4.1 CUDA C++ version - EP Benchmark\n\n");
	printf(" Number of random numbers generated: %15s\n", size);

	verified = FALSE;		

	t1 = A;

	for(i=0; i<MK+1; i++){
		randlc(&t1, t1);
	}

	an = t1;
	gc = 0.0;
	sx = 0.0;
	sy = 0.0;

	for(i=0; i<NQ; i++){
		q[i] = 0.0;
	}

	timer_clear(PROFILING_TOTAL_TIME);
	timer_start(PROFILING_TOTAL_TIME);
	
	skepu::BackendSpec spec{"OpenMP"};
	spec.setCPUThreads(12);
	skepu::setGlobalBackendSpec(spec);
	
	
	auto skel_map = skepu::Map<0>(mapfun);
	skel_map(q0, q1, q2, q3, q4, q5, q6, q7, q8, q9, sx_skepu, sy_skepu, an);
	
	auto skel_reduce = skepu::Reduce([](double lhs, double rhs) -> double { return lhs + rhs; });
	
	q[0] = skel_reduce(q0);
	q[1] = skel_reduce(q1);
	q[2] = skel_reduce(q2);
	q[3] = skel_reduce(q3);
	q[4] = skel_reduce(q4);
	q[5] = skel_reduce(q5);
	q[6] = skel_reduce(q6);
	q[7] = skel_reduce(q7);
	q[8] = skel_reduce(q8);
	q[9] = skel_reduce(q9);
	sx = skel_reduce(sx_skepu);
	sy = skel_reduce(sy_skepu);
	/*
	auto skel_mapreduce = skepu::MapReduce<0>(mapfun, redfun);
	skel_mapreduce.setDefaultSize(NN);
	std::tie(q[0], q[1], q[2], q[3], q[4], q[5], q[6], q[7], q[8], q[9], sx, sy) = skel_mapreduce(an);
*/
	timer_stop(PROFILING_TOTAL_TIME);
	tm = timer_read(PROFILING_TOTAL_TIME);
	
	for(i=0; i<NQ; i++){
		gc+=q[i];
	}				

	nit = 0;
	verified = TRUE;
	if(M == 24){
		sx_verify_value = -3.247834652034740e+3;
		sy_verify_value = -6.958407078382297e+3;
	}else if(M == 25){
		sx_verify_value = -2.863319731645753e+3;
		sy_verify_value = -6.320053679109499e+3;
	}else if(M == 28){
		sx_verify_value = -4.295875165629892e+3;
		sy_verify_value = -1.580732573678431e+4;
	}else if(M == 30){
		sx_verify_value =  4.033815542441498e+4;
		sy_verify_value = -2.660669192809235e+4;
	}else if(M == 32){
		sx_verify_value =  4.764367927995374e+4;
		sy_verify_value = -8.084072988043731e+4;
	}else if(M == 36){
		sx_verify_value =  1.982481200946593e+5;
		sy_verify_value = -1.020596636361769e+5;
	}else if (M == 40){
		sx_verify_value = -5.319717441530e+05;
		sy_verify_value = -3.688834557731e+05;
	}else{
		verified = FALSE;
	}
	if(verified){
		sx_err = fabs((sx - sx_verify_value) / sx_verify_value);
		sy_err = fabs((sy - sy_verify_value) / sy_verify_value);
		verified = ((sx_err <= EPSILON) && (sy_err <= EPSILON));
	}
	Mops = pow(2.0, M+1)/tm/1000000.0;

	printf("\n EP Benchmark Results:\n\n");
	printf(" CPU Time =%10.4f\n", tm);
	printf(" N = 2^%5d\n", M);
	printf(" No. Gaussian Pairs = %15.0f\n", gc);
	printf(" Sums = %25.15e %25.15e\n", sx, sy);
	printf(" Counts: \n");
	for(i=0; i<NQ; i++){
		printf("%3d%15.0f\n", i, q[i]);
	}

#if defined(PROFILING)
	sprintf(gpu_config, "%5s\t%25s\t%25s\t%25s\n", "GPU Kernel", "Threads Per Block", "Time in Seconds", "Time in Percentage");
	strcpy(gpu_config_string, gpu_config);
	sprintf(gpu_config, "%29s\t%25d\t%25f\t%24.2f%%\n", " ep", threads_per_block, timer_read(PROFILING_TOTAL_TIME), (timer_read(PROFILING_TOTAL_TIME)*100/timer_read(PROFILING_TOTAL_TIME)));
	strcat(gpu_config_string, gpu_config);
#endif

	c_print_results((char*)"EP",
			CLASS,
			M+1,
			0,
			0,
			nit,
			tm,
			Mops,
			(char*)"Random numbers generated",
			verified,
			(char*)NPBVERSION,
			(char*)COMPILETIME,
			(char*)COMPILERVERSION,
			(char*)CS1,
			(char*)CS2,
			(char*)CS3,
			(char*)CS4,
			(char*)CS5,
			(char*)CS6,
			(char*)CS7);

	return 0;
}
